@@ -4,7 +4,7 @@
 #include <sys/file.h>
 #include <libc.h>
 #include <mach/mach.h>
-#include "stuff/target_arch.h"
+#include "arch64_32.h"
 #include "stuff/openstep_mach.h"
 #include <mach-o/loader.h>
 #include <mach-o/reloc.h>
@@ -25,10 +25,14 @@
 #ifdef SPARC
 #include <mach-o/sparc/reloc.h>
 #endif
+#ifdef ARM
+#include <mach-o/arm/reloc.h>
+#include "arm_reloc.h"
+#endif
 #if defined(I386) && defined(ARCH64)
 #include <mach-o/x86_64/reloc.h>
 #endif
-#include "stuff/round.h"
+#include "stuff/rnd.h"
 #include "stuff/bytesex.h"
 #include "stuff/errors.h"
 #include "as.h"
@@ -41,6 +45,10 @@
 #include "messages.h"
 #include "xmalloc.h"
 #include "input-scrub.h"
+#include "stuff/write64.h"
+#if defined(I386) && defined(ARCH64)
+#include "i386.h"
+#endif
 #ifdef I860
 #define RELOC_SECTDIFF		I860_RELOC_SECTDIFF
 #define RELOC_LOCAL_SECTDIFF	I860_RELOC_SECTDIFF
@@ -72,6 +80,11 @@
 #define RELOC_LOCAL_SECTDIFF	GENERIC_RELOC_LOCAL_SECTDIFF
 #define RELOC_PAIR		GENERIC_RELOC_PAIR
 #endif
+#ifdef ARM
+#define RELOC_SECTDIFF		ARM_RELOC_SECTDIFF
+#define RELOC_LOCAL_SECTDIFF	ARM_RELOC_SECTDIFF
+#define RELOC_PAIR		ARM_RELOC_PAIR
+#endif
 
 /*
  * These variables are set by layout_symbols() to organize the symbol table and
@@ -86,31 +99,32 @@
  *	strings for local symbols
  */
 /* index to and number of local symbols */
-static unsigned long ilocalsym = 0;
-static unsigned long nlocalsym = 0;
+static uint32_t ilocalsym = 0;
+static uint32_t nlocalsym = 0;
 /* index to, number of and array of sorted externally defined symbols */
-static unsigned long iextdefsym = 0;
-static unsigned long nextdefsym = 0;
+static uint32_t iextdefsym = 0;
+static uint32_t nextdefsym = 0;
 static symbolS **extdefsyms = NULL;
 /* index to, number of and array of sorted undefined symbols */
-static unsigned long iundefsym = 0;
-static unsigned long nundefsym = 0;
+static uint32_t iundefsym = 0;
+static uint32_t nundefsym = 0;
 static symbolS **undefsyms = NULL;
 
-static unsigned long layout_indirect_symbols(
+static uint32_t layout_indirect_symbols(
      void);
 static void layout_symbols(
-    long *symbol_number,
-    long *string_byte_count);
+    int32_t *symbol_number,
+    int32_t *string_byte_count);
 static int qsort_compare(
     const symbolS **sym1,
     const symbolS **sym2);
-static unsigned long nrelocs_for_fix(
+static uint32_t nrelocs_for_fix(
     struct fix *fixP);
-static unsigned long fix_to_relocation_entries(
+static uint32_t fix_to_relocation_entries(
     struct fix *fixP,
-    unsigned long sect_addr,
-    struct relocation_info *riP);
+    uint64_t sect_addr,
+    struct relocation_info *riP,
+    uint32_t debug_section);
 #ifdef I860
 static void
     I860_tweeks(void);
@@ -128,12 +142,13 @@ char *out_file_name)
     segment_command_t		reloc_segment;
     struct symtab_command	symbol_table;
     struct dysymtab_command	dynamic_symbol_table;
-    unsigned long		section_type, *indirect_symbols;
+    uint32_t			section_type;
+    uint32_t			*indirect_symbols;
     isymbolS			*isymbolP;
-    unsigned long		i, j, nsects, nsyms, strsize, nindirectsyms;
+    uint32_t			i, j, nsects, nsyms, strsize, nindirectsyms;
 
     /* locals to fill in section struct fields */
-    unsigned long offset, zero;
+    uint32_t offset, zero;
 
     /* The GAS data structures */
     struct frchain *frchainP, *p;
@@ -141,19 +156,19 @@ char *out_file_name)
     struct frag *fragP;
     struct fix *fixP;
 
-    unsigned long output_size;
+    uint32_t output_size;
     char *output_addr;
     kern_return_t r;
 
     enum byte_sex host_byte_sex;
-    unsigned long reloff, nrelocs;
-    long count;
+    uint32_t reloff, nrelocs;
+    int32_t count;
     char *fill_literal;
-    long fill_size;
-    long num_bytes;
+    int32_t fill_size;
+    int32_t num_bytes;
     char *symbol_name;
     int fd;
-    unsigned long local;
+    uint32_t local;
     struct stat stat_buf;
 
 #ifdef I860
@@ -192,7 +207,7 @@ char *out_file_name)
 	 */
 	nsyms = 0;
 	strsize = 0;
-	layout_symbols((long *)&nsyms, (long *)&strsize);
+	layout_symbols((int32_t *)&nsyms, (int32_t *)&strsize);
 
 	/* fill in the Mach-O header */
 	header.magic = MH_MAGIC_VALUE;
@@ -252,11 +267,16 @@ char *out_file_name)
 	 * zerofill section or the last not-zerofill section.
 	 */
 	for(frchainP = frchain_root; frchainP; frchainP = frchainP->frch_next){
-	    if((frchainP->frch_section.flags & SECTION_TYPE) == S_ZEROFILL)
+	    section_type = frchainP->frch_section.flags & SECTION_TYPE;
+	    if(section_type == S_ZEROFILL ||
+	       section_type == S_THREAD_LOCAL_ZEROFILL)
 		continue;
-	    for(p = frchainP->frch_next; p != NULL; p = p->frch_next)
-		if((p->frch_section.flags & SECTION_TYPE) != S_ZEROFILL)
+	    for(p = frchainP->frch_next; p != NULL; p = p->frch_next){
+		section_type = p->frch_section.flags & SECTION_TYPE;
+		if(section_type != S_ZEROFILL &&
+		   section_type != S_THREAD_LOCAL_ZEROFILL)
 		    break;
+	    }
 	    if(p != NULL)
 		i = p->frch_section.addr - frchainP->frch_section.addr;
 	    else
@@ -268,12 +288,14 @@ char *out_file_name)
 				   frchainP->frch_section.size;
 	}
 	for(frchainP = frchain_root; frchainP; frchainP = frchainP->frch_next){
-	    if((frchainP->frch_section.flags & SECTION_TYPE) != S_ZEROFILL)
+	    section_type = frchainP->frch_section.flags & SECTION_TYPE;
+	    if(section_type != S_ZEROFILL &&
+	       section_type != S_THREAD_LOCAL_ZEROFILL)
 		continue;
 	    reloc_segment.vmsize = frchainP->frch_section.addr +
 				   frchainP->frch_section.size;
 	}
-	offset = round(offset, sizeof(long));
+	offset = rnd(offset, sizeof(int32_t));
 
 	/*
 	 * Count the number of relocation entries for each section.
@@ -288,7 +310,7 @@ char *out_file_name)
 	/*
 	 * Fill in the offset to the relocation entries of the sections.
 	 */
-	offset = round(offset, sizeof(long));
+	offset = rnd(offset, sizeof(int32_t));
 	reloff = offset;
 	nrelocs = 0;
 	for(frchainP = frchain_root; frchainP; frchainP = frchainP->frch_next){
@@ -320,7 +342,7 @@ char *out_file_name)
 	    else{
 		dynamic_symbol_table.nindirectsyms = nindirectsyms;
 		dynamic_symbol_table.indirectsymoff = offset;
-		offset += nindirectsyms * sizeof(unsigned long);
+		offset += nindirectsyms * sizeof(uint32_t);
 	    }
 
 	    dynamic_symbol_table.tocoff = 0;
@@ -350,8 +372,8 @@ char *out_file_name)
 	    symbol_table.stroff = 0;
 	else
 	    symbol_table.stroff = offset;
-	symbol_table.strsize = round(strsize, sizeof(unsigned long));
-	offset += round(strsize, sizeof(unsigned long));
+	symbol_table.strsize = rnd(strsize, sizeof(uint32_t));
+	offset += rnd(strsize, sizeof(uint32_t));
 
 	/*
 	 * The second group of things to do is now with the size of everything
@@ -365,7 +387,7 @@ char *out_file_name)
 	output_size = offset;
 	if((r = vm_allocate(mach_task_self(), (vm_address_t *)&output_addr,
 			    output_size, TRUE)) != KERN_SUCCESS)
-	    as_fatal("can't vm_allocate() buffer for output file of size %lu",
+	    as_fatal("can't vm_allocate() buffer for output file of size %u",
 		     output_size);
 
 	/* put the headers in the output file's buffer */
@@ -508,7 +530,9 @@ char *out_file_name)
 					fixP,
 					frchainP->frch_section.addr,
 					(struct relocation_info *)(output_addr +
-								   offset));
+								   offset),
+				        frchainP->frch_section.flags &
+					  S_ATTR_DEBUG);
 	    }
 	}
 	if(host_byte_sex != md_target_byte_sex)
@@ -548,19 +572,19 @@ char *out_file_name)
 				N_TYPE) == N_ABS)
 				local |= INDIRECT_SYMBOL_ABS;
 			    memcpy(output_addr + offset, (char *)(&local),
-	   			   sizeof(unsigned long));
+	   			   sizeof(uint32_t));
 			}
 			else{
 			    memcpy(output_addr + offset,
 				   (char *)(&isymbolP->isy_symbol->sy_number),
-				   sizeof(unsigned long));
+				   sizeof(uint32_t));
 			}
-			offset += sizeof(unsigned long);
+			offset += sizeof(uint32_t);
 		    }
 		}
 	    }
 	    if(host_byte_sex != md_target_byte_sex){
-		indirect_symbols = (unsigned long *)(output_addr +
+		indirect_symbols = (uint32_t *)(output_addr +
 				    dynamic_symbol_table.indirectsymoff);
 		swap_indirect_symbols(indirect_symbols, nindirectsyms, 
 				      md_target_byte_sex);
@@ -612,7 +636,7 @@ char *out_file_name)
 	}
 	if((fd = open(out_file_name, O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1)
 	    as_fatal("can't create output file: %s", out_file_name);
-	if(write(fd, output_addr, output_size) != (int)output_size)
+	if(write64(fd, output_addr, output_size) != (ssize_t)output_size)
 	    as_fatal("can't write output file");
 	if(close(fd) == -1)
 	    as_fatal("can't close output file");
@@ -624,11 +648,11 @@ char *out_file_name)
  * pointers.  It returns the total count of indirect symbol table entries.
  */
 static
-unsigned long
+uint32_t
 layout_indirect_symbols(void)
 {
     struct frchain *frchainP;
-    unsigned long section_type, total, count, stride;
+    uint32_t section_type, total, count, stride;
     isymbolS *isymbolP;
     symbolS *symbolP;
 
@@ -826,10 +850,10 @@ set_BINCL_checksums()
 static
 void
 layout_symbols(
-long *symbol_number,
-long *string_byte_count)
+int32_t *symbol_number,
+int32_t *string_byte_count)
 {
-    unsigned long i, j;
+    uint32_t i, j;
     symbolS *symbolP;
     symbolS **symbolPP;
     char *name;
@@ -866,7 +890,11 @@ long *string_byte_count)
 		    /* don't keep this symbol */
 		    *symbolPP = symbolP->sy_next;
 		}
-	        else if(flagseen['L'] || (symbolP->sy_type & N_EXT) != 0){
+	        else if(flagseen['L'] || (symbolP->sy_type & N_EXT) != 0
+#if defined(I386) && defined(ARCH64)
+			|| is_section_cstring_literals(symbolP->sy_other)
+#endif
+		){
 		    if((symbolP->sy_type & N_EXT) == 0){
 			nlocalsym++;
 			symbolP->sy_number = *symbol_number;
@@ -933,12 +961,6 @@ long *string_byte_count)
 		    as_bad("Undefined symbol: %s can't be a weak_definition",
 			   symbolP->sy_name);
 		}
-		else if((symbolP->sy_type & N_TYPE) != N_SECT ||
-			is_section_coalesced(symbolP->sy_other) == FALSE){
-		  as_bad("symbol: %s can't be a weak_definition (currently "
-		         "only supported in section of type coalesced)",
-			 symbolP->sy_name);
-		}
 	    }
 	}
 
@@ -1010,7 +1032,7 @@ const symbolS **sym2)
  * specified fix structure.
  */
 static
-unsigned long
+uint32_t
 nrelocs_for_fix(
 struct fix *fixP)
 {
@@ -1061,6 +1083,13 @@ struct fix *fixP)
 	if(fixP->fx_r_type == SPARC_RELOC_HI22 ||
 	   fixP->fx_r_type == SPARC_RELOC_LO10)
 	    return(2);
+#endif
+#ifdef ARM
+	if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_RELOC_HI16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_HI16)
+	    return(2);
 #endif
 	return(1);
 }
@@ -1072,18 +1101,19 @@ struct fix *fixP)
  * placed at riP.
  */
 static
-unsigned long
+uint32_t
 fix_to_relocation_entries(
 struct fix *fixP,
-unsigned long sect_addr,
-struct relocation_info *riP)
+uint64_t sect_addr,
+struct relocation_info *riP,
+uint32_t debug_section)
 {
     struct symbol *symbolP;
-    unsigned long count;
+    uint32_t count;
     struct scattered_relocation_info sri;
-    unsigned long sectdiff;
+    uint32_t sectdiff;
 #ifdef HPPA
-    unsigned long left21, right14;
+    uint32_t left21, right14;
 #endif
 
 	/*
@@ -1099,6 +1129,16 @@ struct relocation_info *riP)
 	memset(riP, '\0', sizeof(struct relocation_info));
 	symbolP = fixP->fx_addsy;
 
+#ifdef ARM
+	/* see arm_reloc.h for the encodings in the low 2 bits */
+	if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_RELOC_HI16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_HI16){
+	    riP->r_length = fixP->fx_r_type & 0x3;
+	}
+	else
+#endif
 	switch(fixP->fx_size){
 	    case 1:
 		riP->r_length = 0;
@@ -1128,19 +1168,39 @@ struct relocation_info *riP)
 	riP->r_pcrel = fixP->fx_pcrel;
 	riP->r_address = fixP->fx_frag->fr_address + fixP->fx_where -
 			 sect_addr;
+#ifdef ARM
+	if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_RELOC_HI16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_HI16){
+	    riP->r_type = ARM_RELOC_HALF;
+	}
+	else
+#endif
 	riP->r_type = fixP->fx_r_type;
 	/*
 	 * For undefined symbols this will be an external relocation entry.
-	 * Or if this is an external coalesced symbol.
+	 * Or if this is an external coalesced symbol or weak symbol.
 	 */
 #if defined(I386) && defined(ARCH64)
-	if (fixP->fx_subsy == NULL && !is_local_symbol(symbolP)) {
+	if(fixP->fx_subsy == NULL &&
+	   (!debug_section || (symbolP->sy_type & N_TYPE) == N_UNDF) &&
+	   (!is_local_symbol(symbolP) ||
+	    ((symbolP->sy_type & N_TYPE) == N_SECT &&
+	     is_section_cstring_literals(symbolP->sy_other)) ) ) {
 #else
 	if((symbolP->sy_type & N_TYPE) == N_UNDF ||
 	   ((symbolP->sy_type & N_EXT) == N_EXT &&
 	    (symbolP->sy_type & N_TYPE) == N_SECT &&
-	    is_section_coalesced(symbolP->sy_other) &&
-	    fixP->fx_subsy == NULL)){
+	    (is_section_coalesced(symbolP->sy_other) ||
+	     (symbolP->sy_desc & N_WEAK_DEF) == N_WEAK_DEF) &&
+	    fixP->fx_subsy == NULL)
+#if defined(I386) && !defined(ARCH64)
+	   ||
+	   ((symbolP->sy_type & N_TYPE) == N_SECT &&
+	    fixP->fx_r_type == GENERIC_RELOC_TLV)
+#endif
+        ){
 #endif
 	    riP->r_extern = 1;
 	    riP->r_symbolnum = symbolP->sy_number;
@@ -1217,9 +1277,17 @@ struct relocation_info *riP)
 		else if(fixP->fx_r_type == SPARC_RELOC_LO10)
 		    sectdiff = SPARC_RELOC_LO10_SECTDIFF;
 		else
+#endif
+#ifdef ARM
+		if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+		   fixP->fx_r_type == ARM_RELOC_HI16 ||
+		   fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+		   fixP->fx_r_type == ARM_THUMB_RELOC_HI16)
+		    sectdiff = ARM_RELOC_HALF_SECTDIFF;
+		else
 #endif
 		{
-		    if(fixP->fx_r_type != 0){
+		    if(fixP->fx_r_type != 0 && fixP->fx_r_type != NO_RELOC){
 			layout_file = fixP->file;
 			layout_line = fixP->line;
 			as_fatal("Internal error: incorrect fx_r_type (%u) for "
@@ -1233,9 +1301,22 @@ struct relocation_info *riP)
 		}
 		memset(&sri, '\0',sizeof(struct scattered_relocation_info));
 		sri.r_scattered = 1;
+#ifdef ARM
+		/* see arm_reloc.h for the encodings in the low 2 bits */
+		if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+		   fixP->fx_r_type == ARM_RELOC_HI16 ||
+		   fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+		   fixP->fx_r_type == ARM_THUMB_RELOC_HI16)
+		    sri.r_length = fixP->fx_r_type & 0x3;
+		else
+#endif
 		sri.r_length    = riP->r_length;
 		sri.r_pcrel     = riP->r_pcrel;
 		sri.r_address   = riP->r_address;
+                if(sri.r_address != riP->r_address)
+		    as_fatal("Section too large, can't encode r_address (0x%x) "
+			     "into 24-bits of scattered relocation entry",
+			     riP->r_address);
 		sri.r_type      = sectdiff;
 		sri.r_value     = symbolP->sy_value;
 		*riP = *((struct relocation_info *)&sri);
@@ -1283,6 +1364,18 @@ struct relocation_info *riP)
 				      fixP->fx_subsy->sy_value +
 				      fixP->fx_offset) >> 10) & 0x3fffff;
 		}
+#endif
+#ifdef ARM
+		else if(sectdiff == ARM_RELOC_HALF_SECTDIFF){
+		    if((sri.r_length & 0x1) == 0x1)
+			sri.r_address = (symbolP->sy_value -
+					 fixP->fx_subsy->sy_value
+					 + fixP->fx_offset) & 0xffff;
+		    else
+			sri.r_address = ((symbolP->sy_value -
+					  fixP->fx_subsy->sy_value +
+					  fixP->fx_offset) >> 16) & 0xffff;
+		}
 #endif
 		*riP = *((struct relocation_info *)&sri);
 		return(2 * sizeof(struct relocation_info));
@@ -1330,6 +1423,10 @@ struct relocation_info *riP)
 		sri.r_length    = riP->r_length;
 		sri.r_pcrel     = riP->r_pcrel;
 		sri.r_address   = riP->r_address;
+                if(sri.r_address != riP->r_address)
+		    as_fatal("Section too large, can't encode r_address (0x%x) "
+			     "into 24-bits of scattered relocation entry",
+			     riP->r_address);
 		sri.r_type      = riP->r_type;
 		sri.r_value     = symbolP->sy_value;
 		*riP = *((struct relocation_info *)&sri);
@@ -1370,9 +1467,25 @@ struct relocation_info *riP)
 #ifdef SPARC
 	if(fixP->fx_r_type == SPARC_RELOC_HI22 ||
 	   fixP->fx_r_type == SPARC_RELOC_LO10)
+#endif
+#ifdef ARM
+	if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_RELOC_HI16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+	   fixP->fx_r_type == ARM_THUMB_RELOC_HI16)
 #endif
 	{
 	    memset(riP, '\0', sizeof(struct relocation_info));
+#ifdef ARM
+	    /* see arm_reloc.h for the encodings in the low 2 bits */
+	    if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+	       fixP->fx_r_type == ARM_RELOC_HI16 ||
+	       fixP->fx_r_type == ARM_THUMB_RELOC_LO16 ||
+	       fixP->fx_r_type == ARM_THUMB_RELOC_HI16){
+		riP->r_length = fixP->fx_r_type & 0x3;
+	    }
+	    else
+#endif
 	    switch(fixP->fx_size){
 		case 1:
 		    riP->r_length = 0;
@@ -1467,6 +1580,15 @@ struct relocation_info *riP)
 		riP->r_address = fixP->fx_value & 0x3ff;
 	    else if (fixP->fx_r_type == SPARC_RELOC_LO10)
 		riP->r_address = (fixP->fx_value >> 10) & 0x3fffff;
+#endif
+#ifdef ARM
+	    riP->r_type = ARM_RELOC_PAIR;
+	    if(fixP->fx_r_type == ARM_RELOC_HI16 ||
+	       fixP->fx_r_type == ARM_THUMB_RELOC_HI16)
+		riP->r_address = 0xffff & fixP->fx_value;
+	    else if(fixP->fx_r_type == ARM_RELOC_LO16 ||
+		    fixP->fx_r_type == ARM_THUMB_RELOC_LO16)
+		riP->r_address = 0xffff & (fixP->fx_value >> 16);
 #endif
 	    count = 2;
 	}
@@ -1484,7 +1606,7 @@ void
 set_default_section_align(
 char *segname,
 char *sectname,
-unsigned long align)
+uint32_t align)
 {
     frchainS *frcP;
 
@@ -1516,7 +1638,8 @@ clear_section_flags(void)
     frchainS *frcP;
 
 	for(frcP = frchain_root; frcP != NULL; frcP = frcP->frch_next)
-	    if(frcP->frch_section.flags != S_ZEROFILL)
+	    if(frcP->frch_section.flags != S_ZEROFILL &&
+	       frcP->frch_section.flags != S_THREAD_LOCAL_ZEROFILL)
 		frcP->frch_section.flags = 0;
 }
 
