@@ -48,8 +48,12 @@ static const struct arch_flag arch_flags[] = {
     /* architecture families */
     { "ppc64",     CPU_TYPE_POWERPC64, CPU_SUBTYPE_POWERPC_ALL },
     { "x86_64",    CPU_TYPE_X86_64, CPU_SUBTYPE_X86_64_ALL },
+    { "x86_64h",   CPU_TYPE_X86_64, CPU_SUBTYPE_X86_64_H },
+    { "arm64",     CPU_TYPE_ARM64,     CPU_SUBTYPE_ARM64_ALL },
     /* specific architecture implementations */
     { "ppc970-64", CPU_TYPE_POWERPC64, CPU_SUBTYPE_POWERPC_970 },
+    { "arm64_32",  CPU_TYPE_ARM64_32,  CPU_SUBTYPE_ARM64_32_V8 },
+    { "arm64e",    CPU_TYPE_ARM64,  CPU_SUBTYPE_ARM64E },
 
 /* 32-bit Mach-O architectures */
 
@@ -62,6 +66,7 @@ static const struct arch_flag arch_flags[] = {
     { "m88k",   CPU_TYPE_MC88000, CPU_SUBTYPE_MC88000_ALL },
     { "i860",   CPU_TYPE_I860,    CPU_SUBTYPE_I860_ALL },
     { "veo",    CPU_TYPE_VEO,     CPU_SUBTYPE_VEO_ALL },
+    { "arm",    CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_ALL },
     /* specific architecture implementations */
     { "ppc601", CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_601 },
     { "ppc603", CPU_TYPE_POWERPC, CPU_SUBTYPE_POWERPC_603 },
@@ -87,9 +92,66 @@ static const struct arch_flag arch_flags[] = {
     { "hppa7100LC", CPU_TYPE_HPPA,  CPU_SUBTYPE_HPPA_7100LC },
     { "veo1",   CPU_TYPE_VEO,     CPU_SUBTYPE_VEO_1 },
     { "veo2",   CPU_TYPE_VEO,     CPU_SUBTYPE_VEO_2 },
+    { "veo3",   CPU_TYPE_VEO,     CPU_SUBTYPE_VEO_3 },
+    { "veo4",   CPU_TYPE_VEO,     CPU_SUBTYPE_VEO_4 },
+    { "armv4t", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V4T},
+    { "armv5",  CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V5TEJ},
+    { "xscale", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_XSCALE},
+    { "armv6",  CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V6 },
+    { "armv6m", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V6M },
+    { "armv7",  CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V7 },
+    { "armv7f", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V7F },
+    { "armv7s", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V7S },
+    { "armv7k", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V7K },
+    { "armv7m", CPU_TYPE_ARM,     CPU_SUBTYPE_ARM_V7M },
+    { "armv7em", CPU_TYPE_ARM,    CPU_SUBTYPE_ARM_V7EM },
+    { "arm64v8",CPU_TYPE_ARM64,   CPU_SUBTYPE_ARM64_V8 },
     { NULL,	0,		  0 }
 };
 
+struct cpu_entry {
+    cpu_type_t 		cputype;
+    enum byte_sex	endian;
+    uint64_t 		staddr;
+    uint32_t		segalign;
+};
+
+/*
+ * The cpu_entries table holds vital statistics for a number of entries in the
+ * arch_info table. Functions such as get_byte_sex_from_flag() are driven off
+ * the contents of this table, making it easier to add new cputypes. Rows in
+ * the table are organized by frequency, with the most common / recent at the
+ * top of the list.
+ *
+ * Note that some historical architectures, such as NeXTSTEP's i386 and
+ * Rhapsody's PPC, are no longer addressible with this design.
+ */
+ 
+static const struct cpu_entry cpu_entries[] = {
+    /* embedded */
+    { CPU_TYPE_ARM64,	    LITTLE_ENDIAN_BYTE_SEX, 0,		       0x4000 },
+    { CPU_TYPE_ARM64_32,    LITTLE_ENDIAN_BYTE_SEX, 0,		       0x4000 },
+    { CPU_TYPE_ARM,	    LITTLE_ENDIAN_BYTE_SEX, 0,		       0x4000 },
+    
+    /* desktop */
+    { CPU_TYPE_X86_64,	    LITTLE_ENDIAN_BYTE_SEX, 0x7fff5fc00000LL,  0x1000 },
+    { CPU_TYPE_I386,	    LITTLE_ENDIAN_BYTE_SEX, 0xc0000000,        0x1000 },
+    { CPU_TYPE_POWERPC,	    BIG_ENDIAN_BYTE_SEX,    0xc0000000,	       0x1000 },
+    { CPU_TYPE_POWERPC64,   BIG_ENDIAN_BYTE_SEX,    0x7ffff00000000LL, 0x1000 },
+    { CPU_TYPE_VEO,	    BIG_ENDIAN_BYTE_SEX,    0xc0000000,	       0x1000 },
+    
+    /* NeXTSTEP / Rhapsody */
+    /*
+    { CPU_TYPE_I386,	    LITTLE_ENDIAN_BYTE_SEX, 0xc0000000,        0x2000 },
+    { CPU_TYPE_POWERPC,	    BIG_ENDIAN_BYTE_SEX,    0xc0000000,	       0x2000 },
+     */
+    { CPU_TYPE_MC680x0,     BIG_ENDIAN_BYTE_SEX,    0x04000000,	       0x2000 },
+    { CPU_TYPE_MC88000,     BIG_ENDIAN_BYTE_SEX,    0xffffe000,	       0x2000 },
+    { CPU_TYPE_SPARC,       BIG_ENDIAN_BYTE_SEX,    0xf0000000,	       0x2000 },
+    { CPU_TYPE_I860,        BIG_ENDIAN_BYTE_SEX,    0,		       0x2000 },
+    { 0 },
+};
+    
 #ifndef RLD
 /*
  * get_arch_from_flag() is passed a name of an architecture flag and returns
@@ -103,7 +165,7 @@ get_arch_from_flag(
 char *name,
 struct arch_flag *arch_flag)
 {
-    unsigned long i;
+    uint32_t i;
 
 	for(i = 0; arch_flags[i].name != NULL; i++){
 	    if(strcmp(arch_flags[i].name, name) == 0){
@@ -142,23 +204,46 @@ get_arch_name_from_types(
 cpu_type_t cputype,
 cpu_subtype_t cpusubtype)
 {
-    unsigned long i;
+    const char *s;
     char *p;
 
+	s = get_arch_name_if_known(cputype, cpusubtype);
+	if (s == NULL) {
+#ifndef RLD
+	    p = savestr("cputype 1234567890 cpusubtype 1234567890");
+	    if(p != NULL)
+		sprintf(p, "cputype %u cpusubtype %u", cputype,
+			cpusubtype & ~CPU_SUBTYPE_MASK);
+#else
+	    /* there is no sprintf() in the rld kernel API's */
+	    p = savestr("cputype ?? cpusubtype ??");
+#endif
+	    s = p;
+	}
+
+	return(s);
+}
+
+/*
+ * get_arch_name_if_known() returns the name of the architecture for the
+ * specified cputype and cpusubtype if known.  If unknown it returns NULL.
+ */
+__private_extern__
+const char *
+get_arch_name_if_known(
+cpu_type_t cputype,
+cpu_subtype_t cpusubtype)
+{
+    uint32_t i;
+
 	for(i = 0; arch_flags[i].name != NULL; i++){
 	    if(arch_flags[i].cputype == cputype &&
-	       arch_flags[i].cpusubtype == cpusubtype)
+	       (arch_flags[i].cpusubtype & ~CPU_SUBTYPE_MASK) ==
+	       (cpusubtype & ~CPU_SUBTYPE_MASK))
 		return(arch_flags[i].name);
 	}
-#ifndef RLD
-	p = savestr("cputype 1234567890 cpusubtype 1234567890");
-	if(p != NULL)
-	    sprintf(p, "cputype %u cpusubtype %u", cputype, cpusubtype);
-#else
-	/* there is no sprintf() in the rld kernel API's */
-	p = savestr("cputype ?? cpusubtype ??");
-#endif
-	return(p);
+
+	return(NULL);
 }
 
 /*
@@ -170,8 +255,11 @@ const struct arch_flag *
 get_arch_family_from_cputype(
 cpu_type_t cputype)
 {
-    unsigned long i;
+    uint32_t i;
 
+	/* arm64 is not to match a family but the specific arm64 arch */
+	if(cputype == CPU_TYPE_ARM64)
+	    return(NULL);
 	for(i = 0; arch_flags[i].name != NULL; i++){
 	    if(arch_flags[i].cputype == cputype)
 		return(arch_flags + i);
@@ -179,6 +267,36 @@ cpu_type_t cputype)
 	return(NULL);
 }
 
+/*
+ * abort_with_unknown_cputype() is a helper function to make calls to abort()
+ * more descriptive in a symbolicated backtrace.
+ */
+static
+void
+abort_with_unknown_cputype(cpu_type_t cputype)
+{
+    abort();
+}
+    
+/*
+ * get_cpu_entry_from_cputype() for the specified cputype. If unknown it
+ * calls abort_with_unknown_cputype().
+ */
+static
+const struct cpu_entry *
+get_cpu_entry_from_cputype(cpu_type_t cputype)
+{
+    for (uint32_t i = 0; cpu_entries[i].cputype != 0; ++i)
+    {
+	if (cpu_entries[i].cputype == cputype)
+	    return cpu_entries + i;
+    }
+    
+    abort_with_unknown_cputype(cputype);
+    
+    return NULL; /* unreachable */
+}
+    
 /*
  * get_byte_sex_from_flag() returns the byte sex of the architecture for the
  * specified cputype and cpusubtype if known.  If unknown it returns
@@ -191,45 +309,12 @@ enum byte_sex
 get_byte_sex_from_flag(
 const struct arch_flag *flag)
 {
-   if(flag->cputype == CPU_TYPE_MC680x0 ||
-      flag->cputype == CPU_TYPE_MC88000 ||
-      flag->cputype == CPU_TYPE_POWERPC ||
-      flag->cputype == CPU_TYPE_POWERPC64 ||
-      flag->cputype == CPU_TYPE_HPPA ||
-      flag->cputype == CPU_TYPE_SPARC ||
-      flag->cputype == CPU_TYPE_I860 ||
-      flag->cputype == CPU_TYPE_VEO)
-        return BIG_ENDIAN_BYTE_SEX;
-    else if(flag->cputype == CPU_TYPE_I386)
-        return LITTLE_ENDIAN_BYTE_SEX;
-    else
-        return UNKNOWN_BYTE_SEX;
+    const struct cpu_entry *entry = get_cpu_entry_from_cputype(flag->cputype);
+    
+    return entry->endian;
 }
 
 #ifndef RLD
-/*
- * get_stack_direction_from_flag() returns the direction the stack grows as
- * either positive (+1) or negative (-1) of the architecture for the
- * specified cputype and cpusubtype if known.  If unknown it returns 0.
- */
-__private_extern__
-long
-get_stack_direction_from_flag(
-const struct arch_flag *flag)
-{
-   if(flag->cputype == CPU_TYPE_MC680x0 ||
-      flag->cputype == CPU_TYPE_MC88000 ||
-      flag->cputype == CPU_TYPE_POWERPC ||
-      flag->cputype == CPU_TYPE_I386 ||
-      flag->cputype == CPU_TYPE_SPARC ||
-      flag->cputype == CPU_TYPE_I860 ||
-      flag->cputype == CPU_TYPE_VEO)
-        return(-1);
-    else if(flag->cputype == CPU_TYPE_HPPA)
-        return(+1);
-    else
-        return(0);
-}
 
 /*
  * get_stack_addr_from_flag() returns the default starting address of the user
@@ -239,80 +324,41 @@ const struct arch_flag *flag)
  * copied here.
  */
 __private_extern__
-unsigned long
+uint64_t
 get_stack_addr_from_flag(
 const struct arch_flag *flag)
 {
-    switch(flag->cputype){
-    case CPU_TYPE_MC680x0:
-	return(0x04000000);
-    case CPU_TYPE_MC88000:
-	return(0xffffe000);
-    case CPU_TYPE_POWERPC:
-    case CPU_TYPE_VEO:
-	return(0xc0000000);
-    case CPU_TYPE_I386:
-	return(0xc0000000);
-    case CPU_TYPE_SPARC:
-	return(0xf0000000);
-    case CPU_TYPE_I860:
-	return(0);
-    case CPU_TYPE_HPPA:
-	return(0xc0000000-0x04000000);
-    default:
-	return(0);
-    }
+    const struct cpu_entry *entry = get_cpu_entry_from_cputype(flag->cputype);
+    
+    return entry->staddr;
 }
 
-/*
- * get_stack_size_from_flag() returns the default size of the userstack.  This
- * should be in the header file <bsd/XXX/vmparam.h> as MAXSSIZ. Since some
- * architectures have come and gone and come back, you can't include all of
- * these headers in one source and some of the constants covered the whole
- * address space the common value of 64meg was chosen.
- */
-__private_extern__
-unsigned long
-get_stack_size_from_flag(
-const struct arch_flag *flag)
-{
-#ifdef __MWERKS__
-    const struct arch_flag *dummy;
-	dummy = flag;
-#endif
-
-    return(64*1024*1024);
-}
-#endif /* !defined(RLD) */
-
 /*
  * get_segalign_from_flag() returns the default segment alignment (page size).
  */
 __private_extern__
-unsigned long
+uint32_t
 get_segalign_from_flag(
 const struct arch_flag *flag)
 {
-	if(flag->cputype == CPU_TYPE_POWERPC ||
-	   flag->cputype == CPU_TYPE_VEO ||
-	   flag->cputype == CPU_TYPE_I386)
-	    return(0x1000); /* 4K */
-	else
-	    return(0x2000); /* 8K */
+    const struct cpu_entry *entry = get_cpu_entry_from_cputype(flag->cputype);
+    
+    return entry->segalign;
 }
 
 /*
- * get_segprot_from_flag() returns the default segment protection.
+ * get_shared_region_size_from_flag() returns the default shared
+ * region size.
  */
 __private_extern__
-vm_prot_t
-get_segprot_from_flag(
+uint32_t
+get_shared_region_size_from_flag(
 const struct arch_flag *flag)
 {
-	if(flag->cputype == CPU_TYPE_I386)
-	    return(VM_PROT_READ | VM_PROT_WRITE);
+	if(flag->cputype == CPU_TYPE_ARM)
+	   return (0x08000000);
 	else
-	    return(VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE);
+	   return (0x10000000);
 }
 
 /*
@@ -330,3 +376,5 @@ cpu_type_t cputype)
 	else
 	    return(FALSE);
 }
+
+#endif /* !defined(RLD) */
