@@ -319,6 +319,29 @@ enum byte_sex target_byte_sex)
 	}
 }
 
+__private_extern__
+void
+swap_fat_arch_64(
+struct fat_arch_64 *fat_archs64,
+uint32_t nfat_arch,
+enum byte_sex target_byte_sex)
+{
+    uint32_t i;
+#ifdef __MWERKS__
+    enum byte_sex dummy;
+        dummy = target_byte_sex;
+#endif
+
+	for(i = 0; i < nfat_arch; i++){
+	    fat_archs64[i].cputype    = SWAP_INT(fat_archs64[i].cputype);
+	    fat_archs64[i].cpusubtype = SWAP_INT(fat_archs64[i].cpusubtype);
+	    fat_archs64[i].offset     = SWAP_LONG_LONG(fat_archs64[i].offset);
+	    fat_archs64[i].size       = SWAP_LONG_LONG(fat_archs64[i].size);
+	    fat_archs64[i].align      = SWAP_INT(fat_archs64[i].align);
+	    fat_archs64[i].reserved   = SWAP_INT(fat_archs64[i].reserved);
+	}
+}
+
 __private_extern__
 void
 swap_mach_header(
@@ -2362,6 +2385,16 @@ enum byte_sex target_byte_sex)
 	cpu->__cpsr = SWAP_INT(cpu->__cpsr);
 }
 
+void
+swap_arm_exception_state64_t(
+arm_exception_state64_t *except,
+enum byte_sex target_byte_sex)
+{
+	except->__far = SWAP_LONG_LONG(except->__far);
+	except->__esr = SWAP_INT(except->__esr);
+	except->__exception = SWAP_INT(except->__exception);
+}
+
 __private_extern__
 void
 swap_ident_command(
@@ -2539,6 +2572,35 @@ enum byte_sex target_byte_sex)
 	ver_cmd->version = SWAP_INT(ver_cmd->version);
 }
 
+__private_extern__
+void
+swap_build_version_command(
+struct build_version_command *bv,
+enum byte_sex target_byte_sex)
+{
+	bv->cmd = SWAP_INT(bv->cmd);
+	bv->cmdsize = SWAP_INT(bv->cmdsize);
+	bv->platform = SWAP_INT(bv->platform);
+	bv->minos = SWAP_INT(bv->minos);
+	bv->sdk = SWAP_INT(bv->sdk);
+	bv->ntools = SWAP_INT(bv->ntools);
+}
+
+__private_extern__
+void
+swap_build_tool_version(
+struct build_tool_version *btv,
+uint32_t ntools,
+enum byte_sex target_byte_sex)
+{
+    uint32_t i;
+
+	for(i = 0; i < ntools; i++){
+	    btv[i].tool = SWAP_INT(btv[i].tool);
+	    btv[i].version = SWAP_INT(btv[i].version);
+	}
+}
+
 __private_extern__
 void swap_rpath_command(
 struct rpath_command *rpath_cmd,
@@ -2630,6 +2692,18 @@ enum byte_sex target_byte_sex)
 	sv->version = SWAP_LONG_LONG(sv->version);
 }
 
+__private_extern__
+void
+swap_note_command(
+struct note_command *nc,
+enum byte_sex target_byte_sex)
+{
+	nc->cmd = SWAP_INT(nc->cmd);
+	nc->cmdsize = SWAP_INT(nc->cmdsize);
+	nc->offset = SWAP_LONG_LONG(nc->offset);
+	nc->size = SWAP_LONG_LONG(nc->size);
+}
+
 __private_extern__
 void
 swap_nlist(
@@ -2693,6 +2767,22 @@ enum byte_sex target_byte_sex)
 	}
 }
 
+__private_extern__
+void
+swap_ranlib_64(
+struct ranlib_64 *ranlibs,
+uint64_t nranlibs,
+enum byte_sex target_byte_sex)
+{
+    uint64_t i;
+
+	for(i = 0; i < nranlibs; i++){
+	    ranlibs[i].ran_un.ran_strx =
+		SWAP_LONG_LONG(ranlibs[i].ran_un.ran_strx);
+	    ranlibs[i].ran_off = SWAP_LONG_LONG(ranlibs[i].ran_off);
+	}
+}
+
 __private_extern__
 void
 swap_relocation_info(
