@@ -171,6 +171,20 @@ struct object *object)
 			"LC_DYLD_INFO load command): ");
 		object->dyld_info = (struct dyld_info_command *)lc;
 	    }
+	    else if(lc->cmd == LC_DYLD_CHAINED_FIXUPS){
+		if(object->dyld_chained_fixups != NULL)
+		    fatal_arch(arch, member, "malformed file (more than one "
+			"LC_DYLD_CHAINED_FIXUPS load command): ");
+		object->dyld_chained_fixups =
+			(struct linkedit_data_command *)lc;
+	    }
+	    else if(lc->cmd == LC_DYLD_EXPORTS_TRIE){
+		if(object->dyld_exports_trie != NULL)
+		    fatal_arch(arch, member, "malformed file (more than one "
+			"LC_DYLD_EXPORTS_TRIE load command): ");
+		object->dyld_exports_trie =
+			(struct linkedit_data_command *)lc;
+	    }
 	    else if(lc->cmd == LC_SEGMENT){
 		sg = (struct segment_command *)lc;
 		if(strcmp(sg->segname, SEG_LINKEDIT) == 0){
@@ -268,7 +282,8 @@ struct object *object)
 		 */
 		dyld_order(arch, member, object);
 	    }
-	    else if(flags & MH_DYLDLINK){
+	    else if(flags & MH_DYLDLINK ||
+		    object->mh_filetype == MH_KEXT_BUNDLE){
 		/*
 		 * This is a file for the dynamic linker (output of ld(1) with
 		 * -output_for_dyld .  That is the relocation entries are split
@@ -355,6 +370,11 @@ struct object *object)
 		   object->seg_bitcode->filesize != offset)
 		    fatal_arch(arch, member, "the __LLVM segment not directly "
 			       "before the "  SEG_LINKEDIT " segment in: ");
+		if(object->seg_bitcode->vmaddr +
+		   object->seg_bitcode->vmsize != object->seg_linkedit->vmaddr)
+		    fatal_arch(arch, member, "the __LLVM segment's vmaddr plus "
+			       "vmsize not directly before the vmaddr of the "
+			       SEG_LINKEDIT " segment in: ");
 	    }
 	}
 	else{
@@ -378,8 +398,15 @@ struct object *object)
 		   object->seg_bitcode64->filesize != offset)
 		    fatal_arch(arch, member, "the __LLVM segment not directly "
 			       "before the "  SEG_LINKEDIT " segment in: ");
+		if(object->seg_bitcode64->vmaddr +
+		   object->seg_bitcode64->vmsize !=
+						object->seg_linkedit64->vmaddr)
+		    fatal_arch(arch, member, "the __LLVM segment's vmaddr plus "
+			       "vmsize not directly before the vmaddr of the "
+			       SEG_LINKEDIT " segment in: ");
 	    }
 	}
+
 	if(object->dyld_info != NULL){
 	    /* dyld_info starts at beginning of __LINKEDIT */
 	    if (object->dyld_info->rebase_off != 0){
@@ -416,6 +443,21 @@ struct object *object)
 		offset = object->dyld_info->rebase_off + 
 			    object->dyld_info->rebase_size;
 	}
+	if(object->dyld_chained_fixups != NULL){
+	    /* dyld_chained_fixups starts at beginning of __LINKEDIT */
+	    if (object->dyld_chained_fixups->dataoff != offset)
+		    order_error(arch, member, "dyld chained fixups "
+			"out of place");
+	    offset = object->dyld_chained_fixups->dataoff +
+	    		object->dyld_chained_fixups->datasize;
+	}
+	if(object->dyld_exports_trie != NULL){
+	    if (object->dyld_exports_trie->dataoff != offset)
+		    order_error(arch, member, "dyld exports trie "
+			"out of place");
+	    offset = object->dyld_exports_trie->dataoff +
+	    		object->dyld_exports_trie->datasize;
+	}
 	if(object->dyst->nlocrel != 0){
 	    if(object->dyst->locreloff != offset)
 		order_error(arch, member, "local relocation entries "
