@@ -30,6 +30,7 @@
 #include "stuff/allocate.h"
 #include "stuff/rnd.h"
 #include "stuff/execute.h"
+#include "stuff/write64.h"
 
 /* used by error routines as the name of the program */
 char *progname = NULL;
@@ -75,7 +76,8 @@ static void strip_bitcode_from_load_commands(
 
 static void leave_only_bitcode_load_commands(
     struct arch *arch,
-    struct object *object);
+    struct object *object,
+    enum bool keeping_plist);
 
 static void reset_pointers_for_object_load_commands(
     struct arch *arch,
@@ -201,7 +203,7 @@ char **envp)
 	    return(EXIT_SUCCESS);
 	}
 
-	writeout(archs, narchs, output, 0777, TRUE, FALSE, FALSE, NULL);
+	writeout(archs, narchs, output, 0777, TRUE, FALSE, FALSE, FALSE, NULL);
 
 	if(errors)
 	    return(EXIT_FAILURE);
@@ -236,6 +238,8 @@ uint32_t narchs)
 
     uint32_t j, offset, size;
     struct object *object;
+    struct ar_hdr h;
+    char size_buf[sizeof(h.ar_size) + 1];
 
 	for(i = 0; i < narchs; i++){
 	    if(archs[i].type == OFILE_ARCHIVE){
@@ -245,7 +249,10 @@ uint32_t narchs)
 			if(check_object(archs + i, archs[i].members + j,
 					object) == FALSE)
 			    return;
-			if(rflag || mflag)
+			if(rflag || mflag ||
+			   (object->mh_filetype != MH_OBJECT &&
+			    object->seg_bitcode == NULL &&
+			    object->seg_bitcode64 == NULL))
 			    strip_bitcode_segment(archs + i,
 						  archs[i].members + j, object);
 			else
@@ -275,15 +282,15 @@ uint32_t narchs)
 			     archs[i].members[j].object->input_sym_info_size +
 			     archs[i].members[j].object->output_sym_info_size,
 			     8);
-			sprintf(archs[i].members[j].ar_hdr->ar_size, "%-*ld",
-			       (int)sizeof(archs[i].members[j].ar_hdr->ar_size),
-			       (long)(size));
+			sprintf(size_buf, "%-*ld",
+			   (int)sizeof(archs[i].members[j].ar_hdr->ar_size),
+			   (long)(size));
 			/*
 			 * This has to be done by hand because sprintf puts a
 			 * null at the end of the buffer.
 			 */
-			memcpy(archs[i].members[j].ar_hdr->ar_fmag, ARFMAG,
-			      (int)sizeof(archs[i].members[j].ar_hdr->ar_fmag));
+			memcpy(archs[i].members[j].ar_hdr->ar_size, size_buf,
+			   (int)sizeof(archs[i].members[j].ar_hdr->ar_size));
 		    }
 		    else{
 			size += archs[i].members[j].unknown_size;
@@ -296,7 +303,10 @@ uint32_t narchs)
 		object = archs[i].object;
 		if(check_object(archs + i, NULL, object) == FALSE)
 		    return;
-		if(rflag || mflag)
+		if(rflag || mflag ||
+		   (object->mh_filetype != MH_OBJECT &&
+		    object->seg_bitcode == NULL &&
+		    object->seg_bitcode64 == NULL))
 		    strip_bitcode_segment(archs + i, NULL, object);
 		else
 		    leave_just_bitcode_segment(archs + i, NULL, object);
@@ -572,6 +582,15 @@ struct object *object)
 		object->seg_linkedit->fileoff += object->seg_bitcode->filesize;
 		start_offset += object->seg_bitcode->filesize;
 	    }
+	    /*
+	     * When removing the bitcode segment move the vmaddr of the
+	     * LINKEDIT down to be contiguous with the previous segment.
+	     * Since it was previously checked in checkout() that the bitcode
+	     * segment is directly before the LINKEDIT segment we just assign
+	     * the vmaddr of the bitcode segment to the LINKEDIT segment
+	     */
+	    if(has_bitcode && rflag)
+		object->seg_linkedit->vmaddr = object->seg_bitcode->vmaddr;
 	}
 	else{
 	    if(object->seg_bitcode64 != NULL){
@@ -620,6 +639,15 @@ struct object *object)
 		    object->seg_bitcode64->filesize;
 		start_offset += object->seg_bitcode64->filesize;
 	    }
+	    /*
+	     * When removing the bitcode segment move the vmaddr of the
+	     * LINKEDIT down to be contiguous with the previous segment.
+	     * Since it was previously checked in checkout() that the bitcode
+	     * segment is directly before the LINKEDIT segment we just assign
+	     * the vmaddr of the bitcode segment to the LINKEDIT segment
+	     */
+	    if(has_bitcode && rflag)
+		object->seg_linkedit64->vmaddr = object->seg_bitcode64->vmaddr;
 	}
 	if(has_bitcode)
 	    some_slice_has_bitcode = TRUE;
@@ -950,7 +978,9 @@ static char fake_string_table[8];
  * leave_just_bitcode_segment() is only called when there is a bitcode segment,
  * and removes everything but that segment contents but leaving the load
  * commands intact but zeros out the sizes and counts to make the resulting
- * Mach-O file valid.
+ * Mach-O file valid.  With the exception of if there is an
+ * (__TEXT,__info_plist) section then that is kept which is a hack to make
+ * code signing work.
  */
 static
 void
@@ -959,12 +989,12 @@ struct arch *arch,
 struct member *member,
 struct object *object)
 {
-    uint32_t i, start_offset, offset, sect_offset;
+    uint32_t i, j, start_offset, offset, sect_offset;
     struct load_command *lc;
-    struct segment_command *sg;
-    struct segment_command_64 *sg64;
-    struct section *s;
-    struct section_64 *s64;
+    struct segment_command *sg, *text;
+    struct segment_command_64 *sg64, *text64;
+    struct section *s, *plist;
+    struct section_64 *s64, *plist64;
 
 	/*
 	 * For MH_OBJECT files, .o files there is no static link editor
@@ -986,7 +1016,8 @@ struct object *object)
 	 * Which is just the size of the old load commands.
 	 *
 	 * To get the bit code segment into the output file we use the
-	 * output_new_content and output_new_content_size.
+	 * output_new_content and output_new_content_size.  And we also use
+	 * this if we have a (__TEXT,__info_plist) section.
 	 *
 	 * Lastly to fake up the output symbolic information will have a 8-byte
 	 * string table of zeros and set the size of all the counts to zero
@@ -994,8 +1025,13 @@ struct object *object)
 	 *
 	 * Also adjust the file offset of the link edit information which is
 	 * where the new fake output symbolic information will start in the
-	 * output file. Which will be right after the load commands.
+	 * output file. Which will be right after the load commands or right
+	 * after the (__TEXT,__info_plist) section.
 	 */
+	text = NULL;
+	text64 = NULL;
+	plist = NULL;
+	plist64 = NULL;
 	if(object->mh != NULL){
 	    start_offset = 0;
 	    lc = object->load_commands;
@@ -1005,8 +1041,18 @@ struct object *object)
 		    if(sg->filesize != 0 && sg->fileoff == 0){
 			s = (struct section *)((char *)sg +
 					       sizeof(struct segment_command));
-			if(sg->nsects > 0)
+			if(sg->nsects > 0){
 			    start_offset = s->offset;
+			    for(j = 0; j < sg->nsects; j++){
+				if(strcmp(s->segname, SEG_TEXT) == 0 &&
+				   strcmp(s->sectname, "__info_plist") == 0 &&
+				   s->size != 0 && plist == NULL){
+				    plist = s;
+				    text = sg;
+				}
+				s++;
+			    }
+			}
 		    }
 		}
 		lc = (struct load_command *)((char *)lc + lc->cmdsize);
@@ -1018,9 +1064,29 @@ struct object *object)
 	    object->object_size -= (object->seg_linkedit->fileoff -
 				    start_offset);
 
-	    object->output_new_content = object->object_addr +
-					 object->seg_bitcode->fileoff;
-	    object->output_new_content_size = object->seg_bitcode->filesize;
+	    if(plist != NULL) {
+		object->output_new_content = object->object_addr +
+					     object->seg_bitcode->fileoff;
+		object->output_new_content_size =
+		    object->seg_bitcode->filesize + plist->size;
+		object->output_new_content =
+		    allocate(object->output_new_content_size);
+		memcpy(object->output_new_content,
+		       object->object_addr + plist->offset,
+		       plist->size);
+		memcpy(object->output_new_content + plist->size,
+		       object->object_addr + object->seg_bitcode->fileoff,
+		       object->seg_bitcode->filesize);
+		plist->offset = start_offset;
+		text->fileoff = plist->offset;
+		text->filesize = plist->size;
+		start_offset += plist->size;
+	    }
+	    else {
+		object->output_new_content = object->object_addr +
+					     object->seg_bitcode->fileoff;
+		object->output_new_content_size = object->seg_bitcode->filesize;
+	    }
 
 	    object->seg_bitcode->fileoff = start_offset;
 	    sect_offset = object->seg_bitcode->fileoff;
@@ -1045,8 +1111,18 @@ struct object *object)
 		    if(sg64->filesize != 0 && sg64->fileoff == 0){
 			s64 = (struct section_64 *)((char *)sg64 +
 					    sizeof(struct segment_command_64));
-			if(sg64->nsects > 0)
+			if(sg64->nsects > 0){
 			    start_offset = s64->offset;
+			    for(j = 0; j < sg64->nsects; j++){
+				if(strcmp(s64->segname, SEG_TEXT) == 0 &&
+				   strcmp(s64->sectname, "__info_plist") == 0 &&
+				   s64->size != 0 && plist64 == NULL){
+				    plist64 = s64;
+				    text64 = sg64;
+				}
+				s64++;
+			    }
+			}
 		    }
 		}
 		lc = (struct load_command *)((char *)lc + lc->cmdsize);
@@ -1058,9 +1134,30 @@ struct object *object)
 	    object->object_size -= (object->seg_linkedit64->fileoff -
 				    start_offset);
 
-	    object->output_new_content = object->object_addr +
-					 object->seg_bitcode64->fileoff;
-	    object->output_new_content_size = object->seg_bitcode64->filesize;
+	    if(plist64 != NULL) {
+		object->output_new_content = object->object_addr +
+					     object->seg_bitcode64->fileoff;
+		object->output_new_content_size =
+		    object->seg_bitcode64->filesize + plist64->size;
+		object->output_new_content =
+		    allocate(object->output_new_content_size);
+		memcpy(object->output_new_content,
+		       object->object_addr + plist64->offset,
+		       plist64->size);
+		memcpy(object->output_new_content + plist64->size,
+		       object->object_addr + object->seg_bitcode64->fileoff,
+		       object->seg_bitcode64->filesize);
+		plist64->offset = start_offset;
+		text64->fileoff = plist64->offset;
+		text64->filesize = plist64->size;
+		start_offset += plist64->size;
+	    }
+	    else {
+		object->output_new_content = object->object_addr +
+					     object->seg_bitcode64->fileoff;
+		object->output_new_content_size =
+		    object->seg_bitcode64->filesize;
+	    }
 
 	    object->seg_bitcode64->fileoff = start_offset;
 	    sect_offset = object->seg_bitcode64->fileoff;
@@ -1218,7 +1315,8 @@ struct object *object)
 	else
 	    object->seg_linkedit64->filesize = object->output_sym_info_size;
 
-	leave_only_bitcode_load_commands(arch, object);
+	leave_only_bitcode_load_commands(arch, object,
+					 plist != NULL || plist64 != NULL);
 }
 
 /*
@@ -1333,13 +1431,17 @@ struct object *object)
  * the LC_SEGMENT or LC_SEGMENT_64 load command from the object's load commands
  * for the bitcode segment.  It only removes the LC_CODE_SIGNATURE and 
  * LC_DYLIB_CODE_SIGN_DRS load commands.  And zeros out all the fields of
- * other load commands to make them valid in the Mach-O file.
+ * other load commands to make them valid in the Mach-O file.  If keeping_plist
+ * is TRUE, then the (__TEXT,__info_plist) section is kept and the caller has
+ * adjusted the __TEXT segment's values and the __info_plist section values,
+ * so they are not changed here.
  */
 static
 void
 leave_only_bitcode_load_commands(
 struct arch *arch,
-struct object *object)
+struct object *object,
+enum bool keeping_plist)
 {
     uint32_t i, j, mh_ncmds, new_ncmds, mh_sizeofcmds, new_sizeofcmds;
     struct load_command *lc1, *lc2, *new_load_commands;
@@ -1348,6 +1450,8 @@ struct object *object)
     struct section *s;
     struct section_64 *s64;
     struct entry_point_command *ep;
+    struct encryption_info_command *encrypt_info;
+    struct encryption_info_command_64 *encrypt_info64;
 
 	/*
 	 * Allocate space for the new load commands and zero it out so any holes
@@ -1378,19 +1482,26 @@ struct object *object)
 		sg = (struct segment_command *)lc1;
 		if(strcmp(sg->segname, "__LLVM") != 0 &&
 		   strcmp(sg->segname, SEG_LINKEDIT) != 0){
-		    sg->vmaddr = 0;
-		    sg->vmsize = 0;
-		    sg->fileoff = 0;
-		    sg->filesize = 0;
+		    if(keeping_plist == FALSE ||
+		       strcmp(sg->segname, SEG_TEXT) != 0){
+			sg->vmaddr = 0;
+			sg->vmsize = 0;
+			sg->fileoff = 0;
+			sg->filesize = 0;
+		    }
 		    s = (struct section *)((char *)sg +
 					   sizeof(struct segment_command));
 		    for(j = 0; j < sg->nsects; j++){
-			s->addr = 0;
-			s->size = 0;
-			s->offset = 0;
-			s->reloff = 0;
-			s->nreloc = 0;
-			s->reserved1 = 0;
+			if(keeping_plist == FALSE ||
+			   strcmp(s->segname, SEG_TEXT) != 0 ||
+			   strcmp(s->sectname, "__info_plist") != 0){
+			    s->addr = 0;
+			    s->size = 0;
+			    s->offset = 0;
+			    s->reloff = 0;
+			    s->nreloc = 0;
+			    s->reserved1 = 0;
+			}
 			s++;
 		    }
 		}
@@ -1403,19 +1514,26 @@ struct object *object)
 		sg64 = (struct segment_command_64 *)lc1;
 		if(strcmp(sg64->segname, "__LLVM") != 0 &&
 		   strcmp(sg64->segname, SEG_LINKEDIT) != 0){
-		    sg64->vmaddr = 0;
-		    sg64->vmsize = 0;
-		    sg64->fileoff = 0;
-		    sg64->filesize = 0;
+		    if(keeping_plist == FALSE ||
+		       strcmp(sg64->segname, SEG_TEXT) != 0){
+			sg64->vmaddr = 0;
+			sg64->vmsize = 0;
+			sg64->fileoff = 0;
+			sg64->filesize = 0;
+		    }
 		    s64 = (struct section_64 *)((char *)sg64 +
 					    sizeof(struct segment_command_64));
 		    for(j = 0; j < sg64->nsects; j++){
-			s64->addr = 0;
-			s64->size = 0;
-			s64->offset = 0;
-			s64->reloff = 0;
-			s64->nreloc = 0;
-			s64->reserved1 = 0;
+			if(keeping_plist == FALSE ||
+			   strcmp(s64->segname, SEG_TEXT) != 0 ||
+			   strcmp(s64->sectname, "__info_plist") != 0){
+			    s64->addr = 0;
+			    s64->size = 0;
+			    s64->offset = 0;
+			    s64->reloff = 0;
+			    s64->nreloc = 0;
+			    s64->reserved1 = 0;
+			}
 			s64++;
 		    }
 		}
@@ -1430,6 +1548,16 @@ struct object *object)
 		    ep = (struct entry_point_command *)lc1;
 		    ep->entryoff = 0;
 		}
+		else if(lc1->cmd == LC_ENCRYPTION_INFO){
+		    encrypt_info = (struct encryption_info_command *)lc1;
+		    encrypt_info->cryptoff = 0;
+		    encrypt_info->cryptsize = 0;
+		}
+		else if(lc1->cmd == LC_ENCRYPTION_INFO_64){
+		    encrypt_info64 = (struct encryption_info_command_64 *)lc1;
+		    encrypt_info64->cryptoff = 0;
+		    encrypt_info64->cryptsize = 0;
+		}
 		memcpy(lc2, lc1, lc1->cmdsize);
 		new_ncmds++;
 		new_sizeofcmds += lc2->cmdsize;
@@ -1541,6 +1669,12 @@ struct object *object)
 	    case LC_CODE_SIGNATURE:
 		object->code_sig_cmd = (struct linkedit_data_command *)lc;
 		break;
+	    case LC_DYLD_CHAINED_FIXUPS:
+		object->dyld_chained_fixups = (struct linkedit_data_command *)lc;
+		break;
+	    case LC_DYLD_EXPORTS_TRIE:
+		object->dyld_exports_trie = (struct linkedit_data_command *)lc;
+		break;
 	    }
 	    lc = (struct load_command *)((char *)lc + lc->cmdsize);
 	}
@@ -1621,8 +1755,8 @@ struct object *object)
 	if((fd = open(input_file, O_WRONLY|O_CREAT, 0600)) < 0)
 	    system_fatal("can't open temporary file: %s", input_file);
 
-	if(write(fd, object->object_addr, object->object_size) !=
-	        object->object_size)
+	if(write64(fd, object->object_addr, object->object_size) !=
+	        (ssize_t)object->object_size)
 	    system_fatal("can't write temporary file: %s", input_file);
 
 	if(close(fd) == -1)
