@@ -41,6 +41,9 @@ struct arch_sign {
 struct arch_sign *arch_signs;
 uint32_t narch_signs = 0;
 
+enum bool rflag = FALSE;
+enum bool pflag = FALSE;
+
 /* used by error routines as the name of the program */
 char *progname = NULL;
 
@@ -62,14 +65,21 @@ static struct linkedit_data_command *add_code_sig_load_command(
     struct member *member,
     struct object *object);
 
+static void strip_LC_CODE_SIGNATURE_commands(
+    struct arch *arch,
+    struct member *member,
+    struct object *object);
+
 /* apple_version is created by the libstuff/Makefile */
 extern char apple_version[];
 char *version = apple_version;
 
+static enum bool cs_alloc_debug;
+
 /*
  * The codesign_allocate(1) tool has the following usage:
  *
- *	codesign_allocate -i oldfile -a arch size ...  -o newfile
+ *	codesign_allocate -i oldfile [-a arch size ...] | [-r] -o newfile
  * 
  * Where the oldfile is a Mach-O file that is input for the dynamic linker
  * and it creates or adds an 
@@ -85,11 +95,31 @@ char **envp)
     struct arch *archs;
     uint32_t narchs;
 
+        /* if CS_ALLOC_DEBUG is set print the complete arguments to stderr */
+        cs_alloc_debug = (NULL != getenv("CS_ALLOC_DEBUG"));
+        if (cs_alloc_debug) {
+            for (i = 0; i < argc; ++i) {
+                if (i < (argc-1)) {
+                    fprintf(stderr, "%s ", argv[i]);
+                } else {
+                    fprintf(stderr, "%s\n", argv[i]);
+                }
+            }
+        }
+
 	progname = argv[0];
 	input = NULL;
 	output = NULL;
 	archs = NULL;
 	narchs = 0;
+	/*
+	 * If this is being run via the symbolic link named codesign_allocate-p
+	 * then set the pflag.
+	 */
+	i = strlen(argv[0]);
+	if(i >= sizeof("codesign_allocate-p") - 1 &&
+	   strcmp(argv[0] + i-2, "-p") == 0)
+	    pflag = TRUE;
 	for(i = 1; i < argc; i++){
 	    if(strcmp(argv[i], "-i") == 0){
 		if(i + 1 == argc){
@@ -183,13 +213,24 @@ char **envp)
 		    i += 3;
 		}
 	    }
+	    else if(strcmp(argv[i], "-r") == 0){
+		rflag = TRUE;
+	    }
+	    else if(strcmp(argv[i], "-p") == 0){
+		pflag = TRUE;
+	    }
 	    else{
 		error("unknown flag: %s", argv[i]);
 		usage();
 	    }
 	}
-	if(input == NULL || output == NULL || narch_signs == 0)
+	if(input == NULL || output == NULL ||
+	   (narch_signs == 0 && rflag == FALSE))
 	    usage();
+	if(rflag && narch_signs != 0){
+	    error("-r flag can't be specified with -a or -A flags");
+	    usage();
+	}
 
 	breakout(input, &archs, &narchs, FALSE);
 	if(errors)
@@ -206,7 +247,7 @@ char **envp)
 		      arch_signs[i].arch_flag.name, arch_signs[i].datasize);
 	}
 
-	writeout(archs, narchs, output, 0777, TRUE, FALSE, FALSE, NULL);
+	writeout(archs, narchs, output, 0777, TRUE, FALSE, FALSE, FALSE, NULL);
 
 	if(errors)
 	    return(EXIT_FAILURE);
@@ -222,8 +263,8 @@ void
 usage(
 void)
 {
-	fprintf(stderr, "Usage: %s -i input [-a <arch> <size>]... "
-		"[-A <cputype> <cpusubtype> <size>]... -o output\n",
+	fprintf(stderr, "Usage: %s -i input [[-a <arch> <size>]... "
+		"[-A <cputype> <cpusubtype> <size>]... | -r] [-p] -o output\n",
 		progname);
 	exit(EXIT_FAILURE);
 }
@@ -239,6 +280,8 @@ struct arch *archs,
 uint32_t narchs)
 {
     uint32_t i, j, offset, size;
+    struct ar_hdr h;
+    char size_buf[sizeof(h.ar_size) + 1];
 
 	for(i = 0; i < narchs; i++){
 	    /*
@@ -271,15 +314,15 @@ uint32_t narchs)
 			size += archs[i].members[j].object->object_size
 			   - archs[i].members[j].object->input_sym_info_size
 			   + archs[i].members[j].object->output_sym_info_size;
-			sprintf(archs[i].members[j].ar_hdr->ar_size, "%-*ld",
-			       (int)sizeof(archs[i].members[j].ar_hdr->ar_size),
-			       (long)(size));
+			sprintf(size_buf, "%-*ld",
+			   (int)sizeof(archs[i].members[j].ar_hdr->ar_size),
+			   (long)(size));
 			/*
 			 * This has to be done by hand because sprintf puts a
 			 * null at the end of the buffer.
 			 */
-			memcpy(archs[i].members[j].ar_hdr->ar_fmag, ARFMAG,
-			      (int)sizeof(archs[i].members[j].ar_hdr->ar_fmag));
+			memcpy(archs[i].members[j].ar_hdr->ar_size, size_buf,
+			   (int)sizeof(archs[i].members[j].ar_hdr->ar_size));
 		    }
 		    else{
 			size += archs[i].members[j].unknown_size;
@@ -295,9 +338,10 @@ uint32_t narchs)
 }
 
 /*
- * setup_code_signature() does the work to add or update the needed
+ * setup_code_signature() does the work to add or update (or remove) the needed
  * LC_CODE_SIGNATURE load command for the specified broken out ofile if it
- * is of one of the architecures specifed with a -a command line options.
+ * is of one of the architecures specifed with a -a command line options.  Or
+ * removes the LC_CODE_SIGNATURE load command if the -r flag is specified.
  */
 static
 void
@@ -306,15 +350,29 @@ struct arch *arch,
 struct member *member,
 struct object *object)
 {
-    uint32_t i;
+    uint32_t i, filetype;
     cpu_type_t cputype;
     cpu_subtype_t cpusubtype;
-    uint32_t flags, linkedit_end;
+    uint32_t flags, linkedit_end, input_sym_info_size_before_code_sig_rnd;
     uint32_t dyld_info_start;
     uint32_t dyld_info_end;
-    uint32_t align_delta;
+    uint32_t align_delta, old_align_delta;
+    uint32_t dataoff, rnd_dataoff;
 
 	linkedit_end = 0;
+	input_sym_info_size_before_code_sig_rnd = 0;
+	if(object->mh != NULL){
+	    cputype = object->mh->cputype;
+	    cpusubtype = object->mh->cpusubtype & ~CPU_SUBTYPE_MASK;
+	    filetype = object->mh->filetype;
+	    flags = object->mh->flags;
+	}
+	else{
+	    cputype = object->mh64->cputype;
+	    cpusubtype = object->mh64->cpusubtype & ~CPU_SUBTYPE_MASK;
+	    filetype = object->mh64->filetype;
+	    flags = object->mh64->flags;
+	}
 	/*
 	 * First set up all the pointers and sizes of the symbolic info.
 	 */
@@ -510,30 +568,78 @@ struct object *object)
 	    }
 	}
 	object->output_sym_info_size = object->input_sym_info_size;
+	input_sym_info_size_before_code_sig_rnd = object->input_sym_info_size;
 	if(object->code_sig_cmd != NULL){
-	    object->input_sym_info_size = rnd(object->input_sym_info_size,
-						16);
+	    /*
+	     * An MH_OBJECT filetype will not have any padding to get the offset
+	     * of an existing code signture to a 16 byte boundary.  To get the
+	     * offset to a specific byte boundary in an MH_OBJECT file the
+	     * string table size before it is adjusted and there is no padding.
+	     */
+	    if(filetype != MH_OBJECT)
+		object->input_sym_info_size = rnd(object->input_sym_info_size,
+						  16);
 	    object->input_sym_info_size += object->code_sig_cmd->datasize;
 	}
 
 	/*
 	 * Now see if one of the -a flags matches this object.
 	 */
-	if(object->mh != NULL){
-	    cputype = object->mh->cputype;
-	    cpusubtype = object->mh->cpusubtype & ~CPU_SUBTYPE_MASK;
-	    flags = object->mh->flags;
-	}
-	else{
-	    cputype = object->mh64->cputype;
-	    cpusubtype = object->mh64->cpusubtype & ~CPU_SUBTYPE_MASK;
-	    flags = object->mh64->flags;
-	}
 	for(i = 0; i < narch_signs; i++){
 	    if(arch_signs[i].arch_flag.cputype == cputype &&
 	       arch_signs[i].arch_flag.cpusubtype == cpusubtype)
 		break;
 	}
+
+	/*
+	 * If the -r flag is specified we remove the code signature and the
+	 * LC_CODE_SIGNATURE load command if any.  We also do this if we have
+	 * an -a flag for this arch and the size is zero.
+	 */
+	if(rflag || (i < narch_signs && arch_signs[i].datasize == 0)){
+	    struct arch_flag arch_flag;
+	    const char *arch_name;
+	    arch_name = get_arch_name_from_types(cputype, cpusubtype);
+	    (void)get_arch_from_flag((char *)arch_name, &arch_flag);
+	    if(i < narch_signs)
+		arch_signs[i].found = TRUE;
+	    /*
+	     * If this has a code signature load command reduce the linkedit by
+	     * the size of that data and the old alignment.  But do not use the
+	     * old data.
+	     */
+	    if(object->code_sig_cmd != NULL){
+		dataoff = object->object_size - object->input_sym_info_size +
+			  input_sym_info_size_before_code_sig_rnd;
+
+		old_align_delta = object->code_sig_cmd->dataoff - dataoff;
+		if(object->seg_linkedit != NULL){
+		    object->seg_linkedit->filesize -=
+			(old_align_delta + object->code_sig_cmd->datasize);
+		    if(object->seg_linkedit->filesize >
+		       object->seg_linkedit->vmsize)
+			object->seg_linkedit->vmsize =
+			    rnd(object->seg_linkedit->filesize,
+		                get_segalign_from_flag(&arch_flag));
+		}
+		else if(object->seg_linkedit64 != NULL){
+		    object->seg_linkedit64->filesize -=
+			old_align_delta;
+		    object->seg_linkedit64->filesize -=
+			object->code_sig_cmd->datasize;
+		    if(object->seg_linkedit64->filesize >
+		       object->seg_linkedit64->vmsize)
+			object->seg_linkedit64->vmsize =
+			    rnd(object->seg_linkedit64->filesize,
+				get_segalign_from_flag(&arch_flag));
+		}
+	    }
+	    object->output_code_sig_data_size = 0;
+	    object->output_code_sig_data = NULL;
+	    strip_LC_CODE_SIGNATURE_commands(arch, member, object);
+	    return;
+	}
+
 	/*
 	 * If we didn't find a matching -a flag then just use the existing
 	 * code signature if any.
@@ -564,9 +670,40 @@ struct object *object)
 	 * the size of that data.  But do not use the old data.
 	 */
 	if(object->code_sig_cmd != NULL){
+	    /*
+	     * To get the code signature data to be page aligned we have to
+	     * determine the old padding incase it was just 16 byte aligned.
+	     * Then determine the new padding alignment and change the string
+	     * table size to add the new padding.
+	     *
+	     * If there is no string table then we just use the existing
+	     * alignment of the old code signature data. 
+	     */
+	    old_align_delta = 0;
+	    align_delta = 0;
+	    if(object->st != NULL) {
+		dataoff = object->object_size - object->input_sym_info_size +
+			  input_sym_info_size_before_code_sig_rnd;
+
+		old_align_delta = object->code_sig_cmd->dataoff - dataoff;
+			
+		if(pflag)
+		    rnd_dataoff = rnd(dataoff, 
+			    get_segalign_from_flag(&arch_signs[i].arch_flag));
+		else
+		    rnd_dataoff = rnd(dataoff, 16);
+		align_delta = rnd_dataoff - dataoff;
+
+		object->code_sig_cmd->dataoff = rnd_dataoff;
+		if(pflag) {
+		    object->output_strings_size_pad = align_delta;
+		    object->st->strsize += object->output_strings_size_pad;
+		}
+	    }
 	    if(object->seg_linkedit != NULL){
 		object->seg_linkedit->filesize +=
-		    arch_signs[i].datasize - object->code_sig_cmd->datasize;
+		    align_delta + arch_signs[i].datasize
+		    - (old_align_delta + object->code_sig_cmd->datasize);
 		if(object->seg_linkedit->filesize >
 		   object->seg_linkedit->vmsize)
 		    object->seg_linkedit->vmsize =
@@ -574,10 +711,14 @@ struct object *object)
 			      get_segalign_from_flag(&arch_signs[i].arch_flag));
 	    }
 	    else if(object->seg_linkedit64 != NULL){
-		object->seg_linkedit64->filesize +=
-		    arch_signs[i].datasize;
+		object->seg_linkedit64->filesize -=
+		    old_align_delta;
 		object->seg_linkedit64->filesize -=
 		    object->code_sig_cmd->datasize;
+		object->seg_linkedit64->filesize +=
+		    align_delta;
+		object->seg_linkedit64->filesize +=
+		    arch_signs[i].datasize;
 		if(object->seg_linkedit64->filesize >
 		   object->seg_linkedit64->vmsize)
 		    object->seg_linkedit64->vmsize =
@@ -589,9 +730,8 @@ struct object *object)
 	    object->output_code_sig_data_size = arch_signs[i].datasize;
 	    object->output_code_sig_data = NULL;
 
-	    object->output_sym_info_size = rnd(object->output_sym_info_size,
-						 16);
-	    object->output_sym_info_size += object->code_sig_cmd->datasize;
+	    object->output_sym_info_size += align_delta +
+					    object->code_sig_cmd->datasize;
 	}
 	/*
 	 * The object does not have a code signature load command we add one.
@@ -618,21 +758,33 @@ struct object *object)
 		      SEG_LINKEDIT " segment", arch->file_name,
 		      arch_signs[i].arch_flag.name);
 
-	    object->code_sig_cmd->dataoff = rnd(linkedit_end, 16);
+	    /*
+	     * If we have a string table and we are padding it so the code
+	     * signature is page aligned determine the string table padding and
+	     * adjust the string table size.  If not just pad to the older 16
+	     * byte alignment.
+	     */
+	    if(object->st != NULL && pflag) {
+		object->code_sig_cmd->dataoff = rnd(linkedit_end,
+			      get_segalign_from_flag(&arch_signs[i].arch_flag));
+		object->output_strings_size_pad =
+		    object->code_sig_cmd->dataoff - linkedit_end;
+		object->st->strsize += object->output_strings_size_pad;
+	    }
+	    else {
+		object->code_sig_cmd->dataoff = rnd(linkedit_end, 16);
+	    }
+	    align_delta = object->code_sig_cmd->dataoff - linkedit_end;
 	    object->output_code_sig_data_size = arch_signs[i].datasize;
 	    object->output_code_sig_data = NULL;
-	    align_delta = object->code_sig_cmd->dataoff - linkedit_end;
 
-	    if(object->output_sym_info_size != 0)
-	        object->output_sym_info_size = rnd(object->output_sym_info_size,
-						   16);
-	    else
-		object->output_sym_info_size = align_delta;
+	    object->output_sym_info_size = align_delta +
+					   object->output_sym_info_size;
 	    object->output_sym_info_size += object->code_sig_cmd->datasize;
 
 	    if(object->seg_linkedit != NULL){
-		object->seg_linkedit->filesize =
-		    rnd(object->seg_linkedit->filesize, 16) +
+		object->seg_linkedit->filesize +=
+		    align_delta +
 		    object->code_sig_cmd->datasize;
 		if(object->seg_linkedit->filesize >
 		   object->seg_linkedit->vmsize)
@@ -641,8 +793,8 @@ struct object *object)
 			      get_segalign_from_flag(&arch_signs[i].arch_flag));
 	    }
 	    else if(object->seg_linkedit64 != NULL){
-		object->seg_linkedit64->filesize =
-		    rnd(object->seg_linkedit64->filesize, 16) +
+		object->seg_linkedit64->filesize +=
+		    align_delta +
 		    object->code_sig_cmd->datasize;
 		if(object->seg_linkedit64->filesize >
 		   object->seg_linkedit64->vmsize)
@@ -692,7 +844,10 @@ struct object *object)
 	 * The size of the new load commands that includes the added code
 	 * signature load command is larger than the existing load commands, so
 	 * see if they can be fitted in before the contents of the first section
-	 * (or segment in the case of a LINKEDIT segment only file).
+	 * (or segment in the case of a LINKEDIT segment only file, in that case
+         * the fileoff is non-zero, as with MH_KEXTBUNDLE it may have a __TEXT
+         * segment with no sections just mapping the headers as those sections
+         * would be in the __TEXT_EXEC segment).
 	 */
 	low_fileoff = UINT_MAX;
 	lc = object->load_commands;
@@ -709,11 +864,12 @@ struct object *object)
 				    S_THREAD_LOCAL_ZEROFILL &&
 			s->offset < low_fileoff)
 			    low_fileoff = s->offset;
-		s++;
+			s++;
 		    }
 		}
 		else{
-		    if(sg->filesize != 0 && sg->fileoff < low_fileoff)
+		    if(sg->fileoff != 0 && sg->filesize != 0 &&
+                       sg->fileoff < low_fileoff)
 			low_fileoff = sg->fileoff;
 		}
 	    }
@@ -733,7 +889,8 @@ struct object *object)
 		    }
 		}
 		else{
-		    if(sg64->filesize != 0 && sg64->fileoff < low_fileoff)
+		    if(sg64->fileoff != 0 && sg64->filesize != 0 &&
+		       sg64->fileoff < low_fileoff)
 			low_fileoff = sg64->fileoff;
 		}
 	    }
@@ -777,3 +934,121 @@ struct object *object)
         }
 	return(code_sig);
 }
+
+/*
+ * strip_LC_CODE_SIGNATURE_commands() is called when -r is specified to remove
+ * any LC_CODE_SIGNATURE load commands from the object's load commands.
+ */
+static
+void
+strip_LC_CODE_SIGNATURE_commands(
+struct arch *arch,
+struct member *member,
+struct object *object)
+{
+    uint32_t i, ncmds, mh_sizeofcmds, sizeofcmds;
+    struct load_command *lc1, *lc2, *new_load_commands;
+    struct segment_command *sg;
+
+	/*
+	 * See if there is an LC_CODE_SIGNATURE load command and if no command
+	 * just return.
+	 */
+	if(object->code_sig_cmd == NULL)
+	    return;
+
+	/*
+	 * Allocate space for the new load commands and zero it out so any holes
+	 * will be zero bytes.
+	 */
+        if(arch->object->mh != NULL){
+            ncmds = arch->object->mh->ncmds;
+	    mh_sizeofcmds = arch->object->mh->sizeofcmds;
+	}
+	else{
+            ncmds = arch->object->mh64->ncmds;
+	    mh_sizeofcmds = arch->object->mh64->sizeofcmds;
+	}
+	new_load_commands = allocate(mh_sizeofcmds);
+	memset(new_load_commands, '\0', mh_sizeofcmds);
+
+	/*
+	 * Copy all the load commands except the LC_CODE_SIGNATURE load commands
+	 * into the allocated space for the new load commands.
+	 */
+	lc1 = arch->object->load_commands;
+	lc2 = new_load_commands;
+	sizeofcmds = 0;
+	for(i = 0; i < ncmds; i++){
+	    if(lc1->cmd != LC_CODE_SIGNATURE){
+		memcpy(lc2, lc1, lc1->cmdsize);
+		sizeofcmds += lc2->cmdsize;
+		lc2 = (struct load_command *)((char *)lc2 + lc2->cmdsize);
+	    }
+	    lc1 = (struct load_command *)((char *)lc1 + lc1->cmdsize);
+	}
+
+	/*
+	 * Finally copy the updated load commands over the existing load
+	 * commands.
+	 */
+	memcpy(arch->object->load_commands, new_load_commands, sizeofcmds);
+	if(mh_sizeofcmds > sizeofcmds){
+		memset((char *)arch->object->load_commands + sizeofcmds, '\0', 
+			   (mh_sizeofcmds - sizeofcmds));
+	}
+	ncmds -= 1;
+        if(arch->object->mh != NULL) {
+            arch->object->mh->sizeofcmds = sizeofcmds;
+            arch->object->mh->ncmds = ncmds;
+        } else {
+            arch->object->mh64->sizeofcmds = sizeofcmds;
+            arch->object->mh64->ncmds = ncmds;
+        }
+	free(new_load_commands);
+
+	/* reset the pointers into the load commands */
+	object->code_sig_cmd = NULL;
+	lc1 = arch->object->load_commands;
+	for(i = 0; i < ncmds; i++){
+	    switch(lc1->cmd){
+	    case LC_SYMTAB:
+		arch->object->st = (struct symtab_command *)lc1;
+	        break;
+	    case LC_DYSYMTAB:
+		arch->object->dyst = (struct dysymtab_command *)lc1;
+		break;
+	    case LC_TWOLEVEL_HINTS:
+		arch->object->hints_cmd = (struct twolevel_hints_command *)lc1;
+		break;
+	    case LC_PREBIND_CKSUM:
+		arch->object->cs = (struct prebind_cksum_command *)lc1;
+		break;
+	    case LC_SEGMENT:
+		sg = (struct segment_command *)lc1;
+		if(strcmp(sg->segname, SEG_LINKEDIT) == 0)
+		    arch->object->seg_linkedit = sg;
+		break;
+	    case LC_SEGMENT_SPLIT_INFO:
+		object->split_info_cmd = (struct linkedit_data_command *)lc1;
+		break;
+	    case LC_FUNCTION_STARTS:
+		object->func_starts_info_cmd =
+					 (struct linkedit_data_command *)lc1;
+		break;
+	    case LC_DATA_IN_CODE:
+		object->data_in_code_cmd =
+				         (struct linkedit_data_command *)lc1;
+		break;
+	    case LC_DYLIB_CODE_SIGN_DRS:
+		object->code_sign_drs_cmd =
+				         (struct linkedit_data_command *)lc1;
+		break;
+	    case LC_LINKER_OPTIMIZATION_HINT:
+		object->link_opt_hint_cmd =
+				         (struct linkedit_data_command *)lc1;
+		break;
+	    }
+	    lc1 = (struct load_command *)((char *)lc1 + lc1->cmdsize);
+	}
+}
