@@ -30,6 +30,7 @@
 #include "stuff/breakout.h"
 #include "stuff/rnd.h"
 #include "stuff/allocate.h"
+#include "stuff/write64.h"
 
 /* used by error routines as the name of the program */
 char *progname = NULL;
@@ -337,7 +338,8 @@ char **envp)
 	    exit(EXIT_FAILURE);
 
 	if(output != NULL)
-	    writeout(archs, narchs, output, 0777, TRUE, FALSE, FALSE, NULL);
+	    writeout(archs, narchs, output, 0777, TRUE, FALSE, FALSE, FALSE,
+		     NULL);
 	else
 	    write_on_input(archs, narchs, input);
 
@@ -416,7 +418,9 @@ char *input)
 	    system_error("can't open input file: %s for writing", input);
 
 	for(i = 0; i < narchs; i++){
-	    if(archs[i].fat_arch != NULL)
+	    if(archs[i].fat_arch64 != NULL)
+		offset = archs[i].fat_arch64->offset;
+	    else if(archs[i].fat_arch != NULL)
 		offset = archs[i].fat_arch->offset;
 	    else
 		offset = 0;
@@ -466,7 +470,7 @@ char *input)
 			fatal("internal error: swap_object_headers() failed");
 	    }
 
-	    if(write(fd, headers, size) != (int)size)
+	    if(write64(fd, headers, size) != (ssize_t)size)
 		system_error("can't write new headers in file: %s", input);
 
 	    free(headers);
@@ -661,6 +665,11 @@ uint32_t *header_size)
     struct rpath_command *rpath1, *rpath2;
     enum bool delete;
 
+	for(i = 0; i < nrpaths; i++)
+	    rpaths[i].found = FALSE;
+	for(i = 0; i < ndelete_rpaths; i++)
+	    delete_rpaths[i].found = FALSE;
+
 	/*
 	 * Make a pass through the load commands and figure out what the new
 	 * size of the the commands needs to be and how much room there is for
