@@ -1138,7 +1138,9 @@ enum bool verbose)
 	}
 
 	if(reloc_found && r_extern == 1){
-	    if((uint32_t)symbols[r_symbolnum].n_un.n_strx >= strings_size)
+	    if(r_symbolnum > nsymbols)
+		name = "bad r_symbolnum offset";
+	    else if((uint32_t)symbols[r_symbolnum].n_un.n_strx >= strings_size)
 		name = "bad string offset";
 	    else
 		name = strings + symbols[r_symbolnum].n_un.n_strx;
